/** @file
 *
 * Copyright (c) 2006 CACE Technologies, Davis (California)
 * All rights reserved.
 *
 * SPDX-License-Identifier: (BSD-3-Clause OR GPL-2.0-only)
 */

#ifndef	_DOT11DECRYPT_WS_H
#define	_DOT11DECRYPT_WS_H

#include "dot11decrypt_system.h"
#include "ws_symbol_export.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

WS_DLL_PUBLIC DOT11DECRYPT_CONTEXT dot11decrypt_ctx;

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _DOT11DECRYPT_WS_H */
